"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import inspect
import subprocess
import os

from inmanta.execute.proxy import DynamicProxy
from inmanta.execute.util import Unknown
from inmanta.ast import Namespace, CompilerException, TypeNotFoundException, RuntimeException
from inmanta.execute.runtime import ExecutionUnit, Resolver, QueueScheduler, ResultVariable
from inmanta.ast.type import TypedList
from inmanta import protocol

from typing import Optional, Callable, List, TYPE_CHECKING, TypeVar, Any, Type

T = TypeVar("T")

if TYPE_CHECKING:
    from inmanta.ast.statements.call import FunctionCall
    from inmanta.ast.statements import ExpressionStatement, DynamicStatement
    from inmanta.compiler import Compiler


class Context(object):
    """
        An instance of this class is used to pass context to the plugin
    """
    __client: Optional[protocol.Client] = None
    __sync_client = None

    @classmethod
    def __get_client(cls) -> protocol.Client:
        if cls.__client is None:
            cls.__client = protocol.Client("compiler")
        return cls.__client

    def __init__(self, resolver: Resolver, queue: QueueScheduler, owner: "FunctionCall", result: ResultVariable) -> None:
        self.resolver = resolver
        self.queue = queue
        self.owner = owner
        self.result = result
        self.compiler = queue.get_compiler()

    def emit_expression(self, stmt: "ExpressionStatement") -> None:
        """
            Add a new statement
        """
        self.owner.copy_location(stmt)
        stmt.normalize(self.resolver)
        reqs = stmt.requires_emit(self.resolver, self.queue)
        ExecutionUnit(self.queue, self.resolver, self.result, reqs, stmt, provides=False)

    def get_resolver(self) -> Resolver:
        return self.resolver

    def get_type(self, name: str):
        """
            Get a type from the configuration model.
        """
        try:
            return self.queue.get_types()[name]
        except KeyError:
            raise TypeNotFoundException(name, self.owner.namespace)

    def get_queue_scheduler(self) -> QueueScheduler:
        return self.queue

    def get_compiler(self) -> "Compiler":
        return self.queue.get_compiler()

    def get_data_dir(self) -> str:
        """
            Get the path to the data dir (and create if it does not exist yet
        """
        data_dir = os.path.join("data", self.owner.function.namespace.get_full_name())

        if not os.path.exists(data_dir):
            os.makedirs(data_dir, exist_ok=True)

        return data_dir

    def get_client(self) -> protocol.Client:
        return self.__class__.__get_client()

    def get_sync_client(self) -> protocol.SyncClient:
        if self.__class__.__sync_client is None:
            self.__class__.__sync_client = protocol.SyncClient("compiler")
        return self.__class__.__sync_client

    def run_sync(self, function: Callable[..., T], timeout: int=5) -> T:
        """
            Execute the async function and return its result. This method takes care of starting and stopping the ioloop. The
            main use for this function is to use the inmanta internal rpc to communicate with the server.

            :param function: The async function to execute. This function should return a yieldable object.
            :param timeout: A timeout for the async function.
            :return: The result of the async call.
            :raises ConnectionRefusedError: When the function timeouts this exception is raised.
        """
        from tornado.ioloop import IOLoop, TimeoutError
        try:
            return IOLoop.current().run_sync(function, timeout)
        except TimeoutError:
            raise ConnectionRefusedError()


class PluginMeta(type):
    """
        A metaclass that registers subclasses in the parent class.
    """
    def __new__(cls, name, bases, dct):
        subclass = type.__new__(cls, name, bases, dct)
        if hasattr(subclass, "__function_name__"):
            cls.add_function(subclass)
        return subclass

    __functions = {}

    @classmethod
    def add_function(cls, plugin_class):
        """
            Add a function plugin class
        """
        name = plugin_class.__function_name__
        ns_parts = str(plugin_class.__module__).split(".")
        ns_parts.append(name)
        if ns_parts[0] != "inmanta_plugins":
            raise Exception(
                "All plugin modules should be loaded in the inmanta_plugins package")

        name = "::".join(ns_parts[1:])
        cls.__functions[name] = plugin_class

    @classmethod
    def get_functions(cls):
        """
            Get all functions that are registered
        """
        return cls.__functions

    @classmethod
    def clear(cls) -> None:
        cls.__functions = {}


class Plugin(object, metaclass=PluginMeta):
    """
        This class models a plugin that can be called from the language.
    """

    def __init__(self, namespace: Namespace) -> None:
        self.ns = namespace

        self._context = -1
        self._return = None

        if hasattr(self.__class__, "__function__"):
            self.arguments = self._load_signature(self.__class__.__function__)
        else:
            self.arguments = []

        self.new_statement = None

    def normalize(self) -> None:
        self.resolver = self.namespace
        self.argtypes = [self.to_type(x[1], self.namespace) for x in self.arguments]
        self.returntype = self.to_type(self._return, self.namespace)

    def _load_signature(self, function):
        """
            Load the signature from the given python function
        """
        arg_spec = inspect.getfullargspec(function)
        if arg_spec.defaults is not None:
            default_start = len(arg_spec.args) - len(arg_spec.defaults)
        else:
            default_start = None

        arguments = []
        for i in range(len(arg_spec.args)):
            arg = arg_spec.args[i]

            if arg not in arg_spec.annotations:
                raise Exception("All arguments of plugin '%s' should be annotated" % function.__name__)

            spec_type = arg_spec.annotations[arg]
            if spec_type == Context:
                self._context = i
            else:
                if default_start is not None and default_start <= i:
                    default_value = arg_spec.defaults[default_start - i]

                    arguments.append((arg, spec_type, default_value))
                else:
                    arguments.append((arg, spec_type))

        if "return" in arg_spec.annotations:
            self._return = arg_spec.annotations["return"]

        return arguments

    def add_argument(self, arg_type, arg_type_name, arg_name, optional=False) -> None:
        """
            Add an argument at the next position, of given type.
        """
        self.arguments.append((arg_type, arg_type_name, arg_name, optional))

    def get_signature(self):
        """
            Generate the signature of this plugin
        """
        arg_list = []
        for arg in self.arguments:
            if len(arg) == 3:
                arg_list.append("%s: %s=%s" % (arg[0], arg[1], str(arg[2])))

            elif len(arg) == 2:
                arg_list.append("%s: %s" % (arg[0], arg[1]))

            else:
                arg_list.append(arg[0])

        args = ", ".join(arg_list)

        if self._return is None:
            return "%s(%s)" % (self.__class__.__function_name__, args)
        return "%s(%s) -> %s" % (self.__class__.__function_name__, args, self._return)

    def to_type(self, arg_type, resolver):
        """
            Convert a string representation of a type to a type
        """
        if arg_type is None:
            return None

        if not isinstance(arg_type, str):
            raise CompilerException("bad annotation in plugin %s::%s, expected str but got %s (%s)" %
                                    (self.ns, self.__class__.__function_name__, type(arg_type), arg_type))

        if arg_type == "any":
            return None

        if arg_type == "list":
            return list

        if arg_type == "expression":
            return None

        if arg_type.endswith("[]"):
            basetypename = arg_type[0:-2]
            basetype = resolver.get_type(basetypename)
            return TypedList(basetype)

        return resolver.get_type(arg_type)

    def _is_instance(self, value: Any, arg_type: Type) -> bool:
        """
            Check if value is of arg_type
        """
        if arg_type is None:
            return True

        if hasattr(arg_type, "validate"):
            return arg_type.validate(value)

        return isinstance(value, arg_type)

    def check_args(self, args: List[Any]) -> bool:
        """
            Check if the arguments of the call match the function signature
        """
        max_arg = len(self.arguments)
        required = len([x for x in self.arguments if len(x) == 2])

        if len(args) < required or len(args) > max_arg:
            raise Exception("Incorrect number of arguments for %s. Expected at least %d, got %d" %
                            (self.get_signature(), required, len(args)))

        for i in range(len(args)):
            if isinstance(args[i], Unknown):
                return False

            if self.arguments[i][0] is not None and not self._is_instance(args[i], self.argtypes[i]):
                raise Exception(("Invalid type for argument %d of '%s', it should be "
                                 "%s and %s given.") % (i + 1, self.__class__.__function_name__,
                                                        self.arguments[i][1], args[i].__class__.__name__))
        return True

    def emit_statement(self) -> "DynamicStatement":
        """
            This method is called to determine if the plugin call pushes a new
            statement
        """
        return self.new_statement

    def is_accept_unknowns(self):
        return self.opts["allow_unknown"]

    def get_variable(self, name, scope):
        """
            Get the given variable
        """
        return DynamicProxy.return_value(self._scope.get_variable(name, scope).value)

    def check_requirements(self) -> None:
        """
            Check if the plug-in has all it requires
        """
        if "bin" in self.opts and self.opts["bin"] is not None:
            for _bin in self.opts["bin"]:
                p = subprocess.Popen(["bash", "-c", "type -p %s" % _bin], stdout=subprocess.PIPE)
                result = p.communicate()

                if len(result[0]) == 0:
                    raise Exception("%s requires %s to be available in $PATH" % (self.__function_name__, _bin))

    def __call__(self, *args):
        """
            The function call itself
        """
        self.check_requirements()
        new_args = []
        for arg in args:
            if isinstance(arg, Context):
                new_args.append(arg)
            elif isinstance(arg, Unknown) and self.is_accept_unknowns():
                new_args.append(arg)
            else:
                new_args.append(DynamicProxy.return_value(arg))

        value = self.call(*new_args)

        value = DynamicProxy.unwrap(value)

        if self.returntype is not None and not isinstance(value, Unknown):
            valid = False
            exception = None

            try:
                valid = (value is None or self._is_instance(value, self.returntype))
            except RuntimeException as e:
                raise e
            except Exception as exp:
                exception = exp

            if not valid:
                msg = ""
                if exception is not None:
                    msg = "\n\tException details: " + str(exception)

                raise Exception("Plugin %s should return value of type %s ('%s' was returned) %s" %
                                (self.__class__.__function_name__, self.returntype, value, msg))

        return value


def plugin(function: Callable=None, commands: List[str]=None, emits_statements: bool=False,
           allow_unknown: bool=False) -> None:  # noqa: H801
    """
        Python decorator to register functions with inmanta as plugin

        :param function: The function to register with inmanta. This is the first argument when it is used as decorator.
        :param commands: A list of command paths that need to be available. Inmanta raises an exception when the command is
                         not available.
        :param emits_statements: Set to true if this plugin emits new statements that the compiler should execute. This is only
                                 required for complex plugins such as integrating a template engine.
        :param allow_unknown: Set to true if this plugin accepts Unknown values as valid input.
    """
    def curry_name(name=None, commands=None, emits_statements=False, allow_unknown=False):
        """
            Function to curry the name of the function
        """
        def call(fnc):
            """
                Create class to register the function and return the function itself
            """

            def wrapper(self, *args):
                """
                    Python will bind the function as method into the class
                """
                return fnc(*args)

            nonlocal name, commands, emits_statements

            if name is None:
                name = fnc.__name__

            dictionary = {}
            dictionary["__module__"] = fnc.__module__
            dictionary["__function_name__"] = name
            dictionary["opts"] = {"bin": commands, "emits_statements": emits_statements, "allow_unknown": allow_unknown}
            dictionary["call"] = wrapper
            dictionary["__function__"] = fnc

            bases = (Plugin,)
            PluginMeta.__new__(PluginMeta, name, bases, dictionary)

            return fnc

        return call

    if function is None:
        return curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif isinstance(function, str):
        return curry_name(function, commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif function is not None:
        fnc = curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)
        return fnc(function)



import subprocess


def test_linkcheck(tmpdir):
    doctrees = tmpdir.join("doctrees")
    htmldir = tmpdir.join("html")
    print(doctrees)
    subprocess.check_call(["sphinx-build", "-W", "-blinkcheck", "-d", str(doctrees), ".", str(htmldir)])


def test_build_docs(tmpdir):
    doctrees = tmpdir.join("doctrees")
    htmldir = tmpdir.join("html")
    subprocess.check_call(["sphinx-build", "-W", "-bhtml", "-d", str(doctrees), ".", str(htmldir)])



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest

from inmanta.ast import TypingException
from inmanta.ast import RuntimeException, DuplicateException
import inmanta.compiler as compiler


def test_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "b" : "b", "c" : 3}
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    b = scope.lookup("b").get_value()
    assert b["a"] == "a"
    assert b["b"] == "b"
    assert b["c"] == 3


def test_dict_collide(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "a" : "b", "c" : 3}
"""
    )

    with pytest.raises(DuplicateException):
        compiler.do_compile()


def test_dict_attr(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Foo:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
end

implement Foo using std::none

a=Foo(bar={})
b=Foo(bar={"a":z})
c=Foo(bar={}, blah={"z":"y"})
z=5
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope

    def map_assert(in_dict, expected):
        for (ek, ev), (k, v) in zip(expected.items(), in_dict.items()):
            assert ek == k
            assert ev == v

    def validate(var, bar, foo, blah):
        e = scope.lookup(var).get_value()
        map_assert(e.get_attribute("bar").get_value(), bar)
        map_assert(e.get_attribute("foo").get_value(), foo)
        map_assert(e.get_attribute("blah").get_value(), blah)

    validate("a", {}, {}, {"a": "a"})
    validate("b", {"a": 5}, {}, {"a": "a"})

    validate("c", {}, {}, {"z": "y"})


def test_dict_attr_type_error(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Foo:
  dict bar
  dict foo = {}
  dict blah = {"a":"a"}
end

implement Foo using std::none

a=Foo(bar=b)
b=Foo(bar={"a":"A"})
"""
    )
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_611_dict_access(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
a = "a"
b = { "a" : a, "b" : "b", "c" : 3}
c=b[a]
d=b["c"]
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    assert scope.lookup("c").get_value() == "a"
    assert scope.lookup("d").get_value() == 3


def test_632_dict_access_2(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
b = { "a" : {"b":"c"}}
c=b["a"]["b"]
"""
    )

    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope
    assert scope.lookup("c").get_value() == "c"


def test_632_dict_access_3(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
b = { "a" : "b"}
c=b["a"]["b"]
"""
    )

    with pytest.raises(TypingException):
        compiler.do_compile()


def test_673_in_dict(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    dict attributes
end

implementation test for Test:

end

implement Test using test when "foo" in self.attributes

Test(attributes={"foo": 42})
"""
    )
    compiler.do_compile()


def test_bad_map_lookup(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        b = {"c" : 3}
        c=b["a"]
        """,
        "key a not found in dict, options are [c] (reported in b['a'] ({dir}/main.cf:3))",
    )



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import sys

from io import StringIO

import inmanta.compiler as compiler


def test_order_of_execution(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
for i in std::sequence(10):
    std::print(i)
end
        """
    )

    saved_stdout = sys.stdout
    try:
        out = StringIO()
        sys.stdout = out
        compiler.do_compile()
        output = out.getvalue().strip()
        assert output == "\n".join([str(x) for x in range(10)])
    finally:
        sys.stdout = saved_stdout


def test_for_error(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        entity A:
            string a = ""
        end
        implement A using std::none
        a = A()
        for i in a:
        end
    """,
        "A for loop can only be applied to lists and relations (reported in For(i) ({dir}/main.cf:7))",
    )


def test_for_error_2(snippetcompiler):
    snippetcompiler.setup_for_error(
        """
        for i in "foo":
        end
    """,
        "A for loop can only be applied to lists and relations (reported in For(i) ({dir}/main.cf:2))",
    )



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import hashlib
import os
import random
import re
import time
import logging
from operator import attrgetter
from itertools import chain
from collections import defaultdict


from inmanta.ast import OptionalValueException, RuntimeException
from inmanta.execute.proxy import DynamicProxy, UnknownException
from inmanta.execute.util import Unknown, NoneValue
from inmanta.export import dependency_manager
from inmanta.plugins import plugin, Context
from inmanta.export import unknown_parameters
from inmanta import resources
from inmanta.module import Project
from inmanta.config import Config


from copy import copy
from inmanta.ast import NotFoundException
from jinja2 import Environment, FileSystemLoader, PrefixLoader
from jinja2.exceptions import UndefinedError
from jinja2.runtime import Undefined
import jinja2


@plugin
def unique_file(prefix: "string", seed: "string", suffix: "string", length: "number"=20) -> "string":
    return prefix + hashlib.md5(seed.encode("utf-8")).hexdigest() + suffix


tcache = {}

engine_cache = None


class JinjaDynamicProxy(DynamicProxy):

    def __init__(self, instance):
        super(JinjaDynamicProxy, self).__init__(instance)

    @classmethod
    def return_value(cls, value):
        if value is None:
            return None

        if isinstance(value, NoneValue):
            return None

        if isinstance(value, Unknown):
            raise UnknownException(value)

        if isinstance(value, (str, tuple, int, float, bool)):
            return copy(value)

        if isinstance(value, DynamicProxy):
            return value

        if hasattr(value, "__len__"):
            return SequenceProxy(value)

        if hasattr(value, "__call__"):
            return CallProxy(value)

        return cls(value)

    def __getattr__(self, attribute):
        instance = self._get_instance()
        if hasattr(instance, "get_attribute"):
            try:
                value = instance.get_attribute(attribute).get_value()
                return JinjaDynamicProxy.return_value(value)
            except (OptionalValueException, NotFoundException):
                return Undefined("variable %s not set on %s" % (attribute, instance), instance, attribute)
        else:
            # A native python object such as a dict
            return getattr(instance, attribute)


class SequenceProxy(JinjaDynamicProxy):

    def __init__(self, iterator):
        JinjaDynamicProxy.__init__(self, iterator)

    def __getitem__(self, key):
        instance = self._get_instance()
        if isinstance(key, str):
            raise RuntimeException(self, "can not get a attribute %s, %s is a list" % (key, self._get_instance()))

        return JinjaDynamicProxy.return_value(instance[key])

    def __len__(self):
        return len(self._get_instance())

    def __iter__(self):
        instance = self._get_instance()

        return IteratorProxy(instance.__iter__())


class CallProxy(JinjaDynamicProxy):
    """
        Proxy a value that implements a __call__ function
    """

    def __init__(self, instance):
        JinjaDynamicProxy.__init__(self, instance)

    def __call__(self, *args, **kwargs):
        instance = self._get_instance()

        return instance(*args, **kwargs)


class IteratorProxy(JinjaDynamicProxy):
    """
        Proxy an iterator call
    """

    def __init__(self, iterator):
        JinjaDynamicProxy.__init__(self, iterator)

    def __iter__(self):
        return self

    def __next__(self):
        i = self._get_instance()
        return JinjaDynamicProxy.return_value(next(i))


class ResolverContext(jinja2.runtime.Context):

    def resolve(self, key):
        resolver = self.parent["{{resolver"]
        try:
            raw = resolver.lookup(key)
            return JinjaDynamicProxy.return_value(raw.get_value())
        except NotFoundException:
            return super(ResolverContext, self).resolve(key)
        except OptionalValueException as e:
            return self.environment.undefined("variable %s not set on %s" % (resolver, key), resolver, key, e)


def _get_template_engine(ctx):
    """
        Initialize the template engine environment
    """
    global engine_cache
    if engine_cache is not None:
        return engine_cache

    loader_map = {}
    loader_map[""] = FileSystemLoader(os.path.join(Project.get().project_path, "templates"))
    for name, module in Project.get().modules.items():
        template_dir = os.path.join(module._path, "templates")
        if os.path.isdir(template_dir):
            loader_map[name] = FileSystemLoader(template_dir)

    # init the environment
    env = Environment(loader=PrefixLoader(loader_map), undefined=jinja2.StrictUndefined)
    env.context_class = ResolverContext

    # register all plugins as filters
    for name, cls in ctx.get_compiler().get_plugins().items():
        def curywrapper(func):
            def safewrapper(*args):
                return JinjaDynamicProxy.return_value(func(*args))
            return safewrapper
        env.filters[name.replace("::", ".")] = curywrapper(cls)

    engine_cache = env
    return env


@plugin("template")
def template(ctx: Context, path: "string"):
    """
        Execute the template in path in the current context. This function will
        generate a new statement that has dependencies on the used variables.
    """
    jinja_env = _get_template_engine(ctx)

    if path in tcache:
        template = tcache[path]
    else:
        template = jinja_env.get_template(path)
        tcache[path] = template

    resolver = ctx.get_resolver()

    try:
        out = template.render({"{{resolver": resolver})
        return out
    except UndefinedError as e:
        raise NotFoundException(ctx.owner, None, e.message)


@dependency_manager
def dir_before_file(model, resources):
    """
        If a file is defined on a host, then make the file depend on its parent directory
    """
    # loop over all resources to find files and dirs
    per_host = defaultdict(list)
    per_host_dirs = defaultdict(list)
    for _id, resource in resources.items():
        if resource.id.get_entity_type() == "std::File" or resource.id.get_entity_type() == "std::Directory":
            per_host[resource.model.host].append(resource)

        if resource.id.get_entity_type() == "std::Directory":
            per_host_dirs[resource.model.host].append(resource)

    # now add deps per host
    for host, files in per_host.items():
        for hfile in files:
            for pdir in per_host_dirs[host]:
                if hfile.path != pdir.path and hfile.path[:len(pdir.path)] == pdir.path:
                    # Make the File resource require the directory
                    hfile.requires.add(pdir)


def get_passwords(pw_file):
    records = {}
    if os.path.exists(pw_file):
        with open(pw_file, "r") as fd:

            for line in fd.readlines():
                line = line.strip()
                if len(line) > 2:
                    i = line.index("=")

                    try:
                        records[line[:i].strip()] = line[i + 1:].strip()
                    except ValueError:
                        pass

    return records


def save_passwords(pw_file, records):
    with open(pw_file, "w+") as fd:
        for key, value in records.items():
            fd.write("%s=%s\n" % (key, value))


@plugin
def generate_password(context: Context, pw_id: "string", length: "number"=20) -> "string":
    """
    Generate a new random password and store it in the data directory of the
    project. On next invocations the stored password will be used.

    :param pw_id: The id of the password to identify it.
    :param length: The length of the password, default length is 20
    """
    data_dir = context.get_data_dir()
    pw_file = os.path.join(data_dir, "passwordfile.txt")

    if "=" in pw_id:
        raise Exception("The password id cannot contain =")

    records = get_passwords(pw_file)

    if pw_id in records:
        return records[pw_id]

    rnd = random.SystemRandom()
    pw = ""
    while len(pw) < length:
        x = chr(rnd.randint(33, 126))
        if re.match("[A-Za-z0-9]", x) is not None:
            pw += x

    # store the new value
    records[pw_id] = pw
    save_passwords(pw_file, records)

    return pw


@plugin
def password(context: Context, pw_id: "string") -> "string":
    """
        Retrieve the given password from a password file. It raises an exception when a password is not found

        :param pw_id: The id of the password to identify it.
    """
    data_dir = context.get_data_dir()
    pw_file = os.path.join(data_dir, "passwordfile.txt")

    if "=" in pw_id:
        raise Exception("The password id cannot contain =")

    records = get_passwords(pw_file)

    if pw_id in records:
        return records[pw_id]

    else:
        raise Exception("Password %s does not exist in file %s" % (pw_id, pw_file))


@plugin("print")
def printf(message: "any"):
    """
        Print the given message to stdout
    """
    print(message)


@plugin
def replace(string: "string", old: "string", new: "string") -> "string":
    return string.replace(old, new)


@plugin
def equals(arg1: "any", arg2: "any", desc: "string"=None):
    """
        Compare arg1 and arg2
    """
    if arg1 != arg2:
        if desc is not None:
            raise AssertionError("%s != %s: %s" % (arg1, arg2, desc))
        else:
            raise AssertionError("%s != %s" % (arg1, arg2))


@plugin("assert")
def assert_function(expression: "bool", message: "string"=""):
    """
        Raise assertion error is expression is false
    """
    if not expression:
        raise AssertionError("Assertion error: " + message)


@plugin
def delay(x: "any") -> "any":
    """
        Delay evaluation
    """
    return x


@plugin
def get(ctx: Context, path: "string") -> "any":
    """
        This function return the variable with given string path
    """
    parts = path.split("::")

    module = parts[0:-1]
    cls_name = parts[-1]

    var = ctx.scope.get_variable(cls_name, module)
    return var.value


@plugin
def select(objects: "list", attr: "string") -> "list":
    """
        Return a list with the select attributes
    """
    r = []
    for obj in objects:
        r.append(getattr(obj, attr))

    return r


@plugin
def item(objects: "list", index: "number") -> "list":
    """
        Return a list that selects the item at index from each of the sublists
    """
    r = []
    for obj in objects:
        r.append(obj[index])

    return r


@plugin
def key_sort(items: "list", key: "any") -> "list":
    """
        Sort an array of object on key
    """
    if isinstance(key, tuple):
        return sorted(items, key=attrgetter(*key))

    return sorted(items, key=attrgetter(key))


@plugin
def timestamp(dummy: "any"=None) -> "number":
    """
        Return an integer with the current unix timestamp

        :param any: A dummy argument to be able to use this function as a filter
    """
    return int(time.time())


@plugin
def capitalize(string: "string") -> "string":
    """
        Capitalize the given string
    """
    return string.capitalize()


@plugin
def type(obj: "any") -> "any":
    value = obj.value
    return value.type().__definition__


@plugin
def sequence(i: "number", start: "number"=0, offset: "number"=0) -> "list":
    """
        Return a sequence of i numbers, starting from zero or start if supplied.
    """
    return list(range(start, int(i) + start - offset))


@plugin
def inlineif(conditional: "bool", a: "any", b: "any") -> "any":
    """
        An inline if
    """
    if conditional:
        return a
    return b


@plugin
def at(objects: "list", index: "number") -> "any":
    """
        Get the item at index
    """
    return objects[int(index)]


@plugin
def attr(obj: "any", attr: "string") -> "any":
    return getattr(obj, attr)


@plugin
def isset(value: "any") -> "bool":
    """
        Returns true if a value has been set
    """
    return value is not None


@plugin
def objid(value: "any") -> "string":
    return str((value._get_instance(), str(id(value._get_instance())), value._get_instance().__class__))


@plugin
def first_of(context: Context, value: "list", type_name: "string") -> "any":
    """
        Return the first in the list that has the given type
    """
    for item in value:
        d = item.type().__definition__
        name = "%s::%s" % (d.namespace, d.name)

        if name == type_name:
            return item

    return None


@plugin
def any(item_list: "list", expression: "expression") -> "bool":
    """
        This method returns true when at least on item evaluates expression
        to true, otherwise it returns false

        :param expression: An expression that accepts one arguments and
            returns true or false
    """
    for item in item_list:
        if expression(item):
            return True
    return False


@plugin
def all(item_list: "list", expression: "expression") -> "bool":
    """
        This method returns false when at least one item does not evaluate
        expression to true, otherwise it returns true

        :param expression: An expression that accepts one argument and
            returns true or false
    """
    for item in item_list:
        if not expression(item):
            return False
    return True


@plugin
def count(item_list: "list") -> "number":
    """
        Returns the number of elements in this list
    """
    return len(item_list)


@plugin
def each(item_list: "list", expression: "expression") -> "list":
    """
        Iterate over this list executing the expression for each item.

        :param expression: An expression that accepts one arguments and
            is evaluated for each item. The returns value of the expression
            is placed in a new list
    """
    new_list = []

    for item in item_list:
        value = expression(item)
        new_list.append(value)

    return new_list


@plugin
def order_by(item_list: "list", expression: "expression"=None, comparator: "expression"=None) -> "list":
    """
        This operation orders a list using the object returned by
        expression and optionally using the comparator function to determine
        the order.

        :param expression: The expression that selects the attributes of the
            items in the source list that are used to determine the order
            of the returned list.

        :param comparator: An optional expression that compares two items.
    """
    expression_cache = {}

    def get_from_cache(item):
        """
            Function that is used to retrieve cache results
        """
        if item in expression_cache:
            return expression_cache[item]
        else:
            data = expression(item)
            expression_cache[item] = data
            return data

    def sort_cmp(item_a, item_b):
        """
            A function that uses the optional expressions to sort item_a list
        """
        if expression is not None:
            a_data = get_from_cache(item_a)
            b_data = get_from_cache(item_b)
        else:
            a_data = item_a
            b_data = item_b

        if comparator is not None:
            return comparator(a_data, b_data)
        else:
            if a_data > b_data:
                return 1
            elif b_data > a_data:
                return -1
            return 0

    # sort
    return sorted(item_list, sort_cmp)


@plugin
def unique(item_list: "list") -> "bool":
    """
        Returns true if all items in this sequence are unique
    """
    seen = set()
    for item in item_list:
        if item in seen:
            return False
        seen.add(item)

    return True


@plugin
def select_attr(item_list: "list", attr: "string") -> "list":
    """
        This query method projects the list onto a new list by transforming
        the list as defined in the expression.
    """
    new_list = []

    for item in item_list:
        new_list.append(lambda x: getattr(x, attr))

    return new_list


@plugin
def select_many(item_list: "list", expression: "expression",
                selector_expression: "expression"=None) -> "list":
    """
        This query method is similar to the select query but it merges
        the results into one list.

        :param expresion: An expression that returns the item that is to be
            included in the resulting list. If that item is a list itself
            it is merged into the result list. The first argument of the
            expression is the item in the source sequence.

        :param selector_expression: This optional arguments allows to
            provide an expression that projects the result of the first
            expression. This selector expression is equivalent to what the
            select method expects. If the returned item of expression is
            not a list this expression is not applied.
    """
    new_list = []

    for item in item_list:
        result = expression(item)

        if not hasattr(result, "__iter__"):
            new_list.append(result)
        else:
            if selector_expression:
                for result_item in result:
                    new_list.append(selector_expression(result_item))
            else:
                new_list.extend(result)

    return new_list


@plugin
def where(item_list: "list", expression: "expression") -> "list":
    """
        This query method selects the items in the list that evaluate the
        expression to true.

        :param expression: An expression that returns true or false
            to determine if an item from the list is included. The first
            argument of the expression is the item that is to be evaluated.
            The second optional argument is the index of the item in the
            list.
    """
    new_list = []
    for index in range(len(item_list)):
        item = item_list[index]

        if expression(item):
            new_list.append(item)

    return new_list


@plugin
def where_compare(item_list: "list", expr_list: "list") -> "list":
    """
        This query selects items in a list but uses the tupples in expr_list
        to select the items.

        :param expr_list: A list of tupples where the first item is the attr
            name and the second item in the tupple is the value
    """
    new_list = []

    new_expr_list = []
    for i in range(0, len(expr_list), 2):
        new_expr_list.append((expr_list[i], expr_list[i + 1]))

    for index in range(len(item_list)):
        item = item_list[index]

        for attr, value in new_expr_list:
            if getattr(item, attr) == value:
                new_list.append(item)

    return new_list


@plugin
def flatten(item_list: "list") -> "list":
    """
        Flatten this list
    """
    return list(chain.from_iterable(item_list))


@plugin
def split(string_list: "string", delim: "string") -> "list":
    """
        Split the given string into a list

        :param string_list: The list to split into parts
        :param delim: The delimeter to split the text by
    """
    return string_list.split(delim)


def determine_path(ctx, module_dir, path):
    """
        Determine the real path based on the given path
    """
    parts = path.split(os.path.sep)

    modules = Project.get().modules

    if parts[0] == "":
        module_path = Project.get().project_path
    elif parts[0] not in modules:
        raise Exception("Module %s does not exist for path %s" %
                        (parts[0], path))
    else:
        module_path = modules[parts[0]]._path

    return os.path.join(module_path, module_dir, os.path.sep.join(parts[1:]))


def get_file_content(ctx, module_dir, path):
    """
        Get the contents of a file
    """
    filename = determine_path(ctx, module_dir, path)

    if filename is None:
        raise Exception("%s does not exist" % path)

    if not os.path.isfile(filename):
        raise Exception("%s isn't a valid file (%s)" % (path, filename))

    file_fd = open(filename, 'r')
    if file_fd is None:
        raise Exception("Unable to open file %s" % filename)

    content = file_fd.read()
    file_fd.close()

    return content


@plugin
def source(ctx: Context, path: "string") -> "string":
    """
        Return the textual contents of the given file
    """
    return get_file_content(ctx, 'files', path)


@plugin
def file(ctx: Context, path: "string") -> "string":
    """
        Return the textual contents of the given file
    """
    filename = determine_path(ctx, 'files', path)
    any
    if filename is None:
        raise Exception("%s does not exist" % path)

    if not os.path.isfile(filename):
        raise Exception("%s isn't a valid file" % path)

    return "imp-module-source:file://" + os.path.abspath(filename)


@plugin
def familyof(member: "std::OS", family: "string") -> "bool":
    """
        Determine if member is a member of the given operating system family
    """
    if member.name == family:
        return True

    parent = member
    try:
        while parent.family is not None:
            if parent.name == family:
                return True

            parent = parent.family
    except OptionalValueException:
        pass

    return False


@plugin
def getfact(context: Context, resource: "any", fact_name: "string", default_value: "any"=None) -> "any":
    """
        Retrieve a fact of the given resource
    """
    resource_id = resources.to_id(resource)
    if resource_id is None:
        raise Exception("Facts can only be retreived from resources.")

    # Special case for unit testing and mocking
    if hasattr(context.compiler, "refs") and "facts" in context.compiler.refs:
        if resource_id in context.compiler.refs["facts"] and fact_name in context.compiler.refs["facts"][resource_id]:
            return context.compiler.refs["facts"][resource_id][fact_name]

        fact_value = Unknown(source=resource)
        unknown_parameters.append({"resource": resource_id, "parameter": fact_name, "source": "fact"})

        if default_value is not None:
            return default_value
        return fact_value
    # End special case

    fact_value = None
    try:
        client = context.get_client()

        env = Config.get("config", "environment", None)
        if env is None:
            raise Exception("The environment of this model should be configured in config>environment")

        def call():
            return client.get_param(tid=env, id=fact_name, resource_id=resource_id)

        result = context.run_sync(call)

        if result.code == 200:
            fact_value = result.result["parameter"]["value"]
        else:
            logging.getLogger(__name__).info("Param %s of resource %s is unknown", fact_name, resource_id)
            fact_value = Unknown(source=resource)
            unknown_parameters.append({"resource": resource_id, "parameter": fact_name, "source": "fact"})

    except ConnectionRefusedError:
        logging.getLogger(__name__).warning("Param %s of resource %s is unknown because connection to server was refused",
                                            fact_name, resource_id)
        fact_value = Unknown(source=resource)
        unknown_parameters.append({"resource": resource_id, "parameter": fact_name, "source": "fact"})

    if isinstance(fact_value, Unknown) and default_value is not None:
        return default_value

    return fact_value


@plugin
def environment() -> "string":
    """
        Return the environment id
    """
    env = str(Config.get("config", "environment", None))

    if env is None:
        raise Exception("The environment of this model should be configured in config>environment")

    return str(env)


@plugin
def environment_name(ctx: Context) -> "string":
    """
        Return the name of the environment (as defined on the server)
    """
    env = environment()

    def call():
        return ctx.get_client().get_environment(id=env)
    result = ctx.run_sync(call)
    if result.code != 200:
        return Unknown(source=env)
    return result.result["environment"]["name"]


@plugin
def environment_server(ctx: Context) -> "string":
    """
        Return the address of the management server
    """
    client = ctx.get_client()
    server_url = client._transport_instance._get_client_config()
    match = re.search("^http[s]?://([^:]+):", server_url)
    if match is not None:
        return match.group(1)
    return Unknown(source=server_url)


@plugin
def is_set(obj: "any", attribute: "string") -> "bool":
    try:
        getattr(obj, attribute)
    except Exception:
        return False
    return True


@plugin
def server_ca() -> "string":
    filename = Config.get("compiler_rest_transport", "ssl_ca_cert_file", "")
    if filename == "":
        return ""
    if filename is None:
        raise Exception("%s does not exist" % filename)

    if not os.path.isfile(filename):
        raise Exception("%s isn't a valid file" % filename)

    file_fd = open(filename, 'r')
    if file_fd is None:
        raise Exception("Unable to open file %s" % filename)

    content = file_fd.read()
    return content


@plugin
def server_token(context: Context, client_types: "string[]" = ["agent"]) -> "string":
    token = Config.get("compiler_rest_transport", "token", "")
    if token == "":
        return ""

    # Request a new token for this agent
    token = ""
    try:
        client = context.get_client()

        env = Config.get("config", "environment", None)
        if env is None:
            raise Exception("The environment of this model should be configured in config>environment")

        def call():
            return client.create_token(tid=env, client_types=list(client_types), idempotent=True)

        result = context.run_sync(call)

        if result.code == 200:
            token = result.result["token"]
        else:
            logging.getLogger(__name__).warning("Unable to get a new token")
            raise Exception("Unable to get a valid token")
    except ConnectionRefusedError:
        logging.getLogger(__name__).exception("Unable to get a new token")
        raise Exception("Unable to get a valid token")

    return token


@plugin
def server_port() -> "number":
    return Config.get("compiler_rest_transport", "port", 8888)


@plugin
def get_env(name: "string", default_value: "string"=None) -> "string":
    env = os.environ
    if name in env:
        return env[name]
    elif default_value is not None:
        return default_value
    else:
        return Unknown(source=name)


@plugin
def get_env_int(name: "string", default_value: "number"=None) -> "number":
    env = os.environ
    if name in env:
        return int(env[name])
    elif default_value is not None:
        return default_value
    else:
        return Unknown(source=name)


@plugin
def is_instance(ctx: Context, obj: "any", cls: "string") -> "bool":
    t = ctx.get_type(cls)
    try:
        t.validate(obj._get_instance())
    except RuntimeException:
        return False
    return True


@plugin("if")
def inlineif(expr: "bool", true_value: "any", false_value: "any") -> "any":
    """
        An inline if expression that can also handle nullable booleans

        :attr expr: The if expression to determine the true or false.
        :attr true_value: The value when expr is true
        :attr false_value: The value when the expr is not true
    """
    if expr:
        return true_value
    return false_value



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import inspect
import json
import uuid
from datetime import datetime
import enum

from tornado import gen, escape
from inmanta import const
from inmanta.protocol import common, exceptions
from inmanta import config as inmanta_config

from typing import Any, Dict, List, Optional, Tuple, TYPE_CHECKING, cast, Mapping, Generator  # noqa: F401

JsonDict = Dict[str, Any]

LOGGER: logging.Logger = logging.getLogger(__name__)
INMANTA_MT_HEADER = "X-Inmanta-tid"
CONTENT_TYPE = "Content-Type"
JSON_CONTENT = "application/json"

"""

RestServer  => manages tornado/handlers, marshalling, dispatching, and endpoints

ServerSlice => contributes handlers and methods

ServerSlice.server [1] -- RestServer.endpoints [1:]

"""


def authorize_request(
    auth_data: Dict[str, str], metadata: Dict[str, str], message: Dict[str, Any], config: common.UrlMethod
) -> None:
    """
        Authorize a request based on the given data
    """
    if auth_data is None:
        return

    # Enforce environment restrictions
    env_key: str = const.INMANTA_URN + "env"
    if env_key in auth_data:
        if env_key not in metadata:
            raise exceptions.UnauthorizedException("The authorization token is scoped to a specific environment.")

        if metadata[env_key] != "all" and auth_data[env_key] != metadata[env_key]:
            raise exceptions.UnauthorizedException("The authorization token is not valid for the requested environment.")

    # Enforce client_types restrictions
    ok: bool = False
    ct_key: str = const.INMANTA_URN + "ct"
    for ct in auth_data[ct_key]:
        if ct in config.properties.client_types:
            ok = True

    if not ok:
        raise exceptions.UnauthorizedException(
            "The authorization token does not have a valid client type for this call."
            + " (%s provided, %s expected" % (auth_data[ct_key], config.properties.client_types)
        )


class CallArguments(object):
    """
        This class represents the call arguments for a method call.
    """

    def __init__(
        self, properties: common.MethodProperties, message: Dict[str, Optional[Any]], request_headers: Mapping[str, str]
    ) -> None:
        """
            :param method_config: The method configuration that contains the metadata and functions to call
            :param message: The message recieved by the RPC call
            :param request_headers: The headers received by the RPC call
        """
        self._properties = properties
        self._message = message
        self._request_headers = request_headers
        self._argspec = inspect.getfullargspec(self._properties.function)

        self._call_args: Dict[str, Any] = {}
        self._headers: Dict[str, str] = {}
        self._metadata: Dict[str, Any] = {}

        self._processed: bool = False

    @property
    def call_args(self) -> Dict[str, Any]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._call_args

    @property
    def headers(self) -> Dict[str, str]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._headers

    @property
    def metadata(self) -> Dict[str, Any]:
        if not self._processed:
            raise Exception("Process call first before accessing property")

        return self._metadata

    def _map_headers(self, arg: str) -> Optional[Any]:
        if arg not in self._properties.arg_options:
            return None

        opts = self._properties.arg_options[arg]

        if opts.header is None:
            return None

        value = self._request_headers[opts.header]
        if opts.reply_header:
            self._headers[opts.header] = value

        return value

    def _process_typing(self, arg: str, value: Optional[Any]) -> Optional[Any]:
        """
            Validate and coerce if required
            :param arg: The name of the arugment
            :param value: The current value of the argument
            :return: The processed value of the argument
        """
        if arg not in self._argspec.annotations:
            return value

        if value is None:
            return value

        arg_type = self._argspec.annotations[arg]
        if isinstance(value, arg_type):
            return value

        try:
            if arg_type == datetime:
                return datetime.strptime(value, const.TIME_ISOFMT)

            elif issubclass(arg_type, enum.Enum):
                return arg_type[value]

            elif arg_type == bool:
                return inmanta_config.is_bool(value)

            else:
                return arg_type(value)

        except (ValueError, TypeError):
            error_msg = "Invalid type for argument %s. Expected %s but received %s, %s" % (
                arg,
                arg_type,
                value.__class__,
                value,
            )
            LOGGER.exception(error_msg)
            raise exceptions.BadRequest(error_msg)

    def get_default_value(self, arg_name: str, arg_position: int, default_start: int) -> Optional[Any]:
        """
            Get a default value for an argument
        """
        if default_start >= 0 and (arg_position - default_start) < len(self._argspec.defaults):
            return self._argspec.defaults[arg_position - default_start]
        else:
            raise exceptions.BadRequest("Invalid request. Field '%s' is required." % arg_name)

    @gen.coroutine
    def _run_getters(self, arg: str, value: Optional[Any]) -> Optional[Any]:
        """
            Run ant available getters on value
        """
        if arg not in self._properties.arg_options or self._properties.arg_options[arg].getter is None:
            return value

        try:
            value = yield self._properties.arg_options[arg].getter(value, self._metadata)
            return value
        except Exception as e:
            LOGGER.exception("Failed to use getter for arg %s", arg)
            raise e

    @gen.coroutine
    def process(self) -> None:
        """
            Process the message
        """
        args: List[str] = list(self._argspec.args)

        if "self" in args:
            args.remove("self")

        all_fields = set(self._message.keys())  # Track all processed fields to warn user
        defaults_start: int = -1
        if self._argspec.defaults is not None:
            defaults_start = len(args) - len(self._argspec.defaults)

        for i, arg in enumerate(args):
            # get value from headers, defaults or message
            value = self._map_headers(arg)
            if value is None:
                if arg in self._message:
                    value = self._message[arg]
                    all_fields.remove(arg)

                else:  # get default value
                    value = self.get_default_value(arg, i, defaults_start)

            # validate type
            value = self._process_typing(arg, value)

            # run getters
            value = yield self._run_getters(arg, value)

            self._call_args[arg] = value

        # discard session handling data
        if self._properties.agent_server and "sid" in all_fields:
            all_fields.remove("sid")

        if len(all_fields) > 0 and self._argspec.varkw is None:
            raise exceptions.BadRequest(
                "request contains fields %s that are not declared in method and no kwargs argument is provided." % all_fields
            )

        self._processed = True


# Shared
class RESTBase(object):
    """
        Base class for REST based client and servers
    """

    _id: str

    @property
    def id(self) -> str:
        return self._id

    def _decode(self, body: str) -> Optional[JsonDict]:
        """
            Decode a response body
        """
        result = None
        if body is not None and len(body) > 0:
            result = cast(JsonDict, json.loads(escape.to_basestring(body)))

        return result

    def validate_sid(self, sid: uuid.UUID) -> bool:
        raise NotImplementedError()

    @gen.coroutine
    def _execute_call(
        self,
        kwargs: Dict[str, str],
        http_method: str,
        config: common.UrlMethod,
        message: Dict[str, Any],
        request_headers: Mapping[str, str],
        auth=None,
    ) -> Generator[Any, Any, common.Response]:

        headers: Dict[str, str] = {}
        try:
            if kwargs is None or config is None:
                raise Exception("This method is unknown! This should not occur!")

            # create message that contains all arguments (id, query args and body)
            if "id" in kwargs and (message is None or "id" not in message):
                message["id"] = kwargs["id"]

            # validate message against config
            if config.properties.id and "id" not in message:
                raise exceptions.BadRequest("the request should contain an id in the url.")

            if config.properties.validate_sid:
                if "sid" not in message:
                    raise exceptions.BadRequest("this is an agent to server call, it should contain an agent session id")

                elif not self.validate_sid(uuid.UUID(message["sid"])):
                    raise exceptions.BadRequest("the sid %s is not valid." % message["sid"])

            arguments = CallArguments(config.properties, message, request_headers)
            yield arguments.process()
            authorize_request(auth, arguments.metadata, arguments.call_args, config)

            # rename arguments if handler requests this
            call_args = arguments.call_args
            if hasattr(config.handler, "__protocol_mapping__"):
                for k, v in config.handler.__protocol_mapping__.items():
                    if v in call_args:
                        call_args[k] = call_args[v]
                        del call_args[v]

            LOGGER.debug(
                "Calling method %s(%s)",
                config.handler,
                ", ".join(["%s='%s'" % (name, common.shorten(str(value))) for name, value in arguments.call_args.items()]),
            )

            result = yield config.handler(**arguments.call_args)

            if result is None:
                raise exceptions.BadRequest(
                    "Handlers for method calls should at least return a status code. %s on %s"
                    % (config.method_name, config.endpoint)
                )

            reply = None
            if isinstance(result, tuple):
                if len(result) == 2:
                    code, reply = result
                else:
                    raise exceptions.BadRequest("Handlers for method call can only return a status code and a reply")

            else:
                code = result

            if reply is not None:
                if config.properties.reply:
                    LOGGER.debug("%s returned %d: %s", config.method_name, code, common.shorten(str(reply), 70))
                    return common.Response(body=reply, headers=headers, status_code=code)

                else:
                    LOGGER.warning("Method %s returned a result although it is has not reply!")

            return common.Response(headers=headers, status_code=code)

        except exceptions.BaseException:
            LOGGER.exception("")
            raise

        except Exception as e:
            LOGGER.exception("An exception occured during the request.")
            raise exceptions.ServerError(str(e.args))



from inmanta.plugins import plugin


@plugin
def list_id(listin: "test::Item[]") -> "test::Item[]":
    return listin


@plugin
def unwrap_list(listin: "test::Item[]") -> "test::Item[]":
    return [x for x in listin]


@plugin
def makelist(ins: "string") -> "string[]":
    return [ins]


@plugin
def badtype(listin: "test::Item[]") -> "test::Item[]":
    return ["a", "b", "c"]



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import ssl
import uuid

from typing import Optional, Dict, Any, List, Generator, NoReturn

import tornado
from tornado import gen, httpserver, web, routing

import inmanta.protocol.endpoints
from inmanta import config as inmanta_config, const
from inmanta.protocol import exceptions, common
from inmanta.protocol.rest import LOGGER, CONTENT_TYPE, JSON_CONTENT, RESTBase


NoneGen = Generator[Any, Any, NoReturn]


class RESTHandler(tornado.web.RequestHandler):
    """
        A generic class use by the transport
    """

    def initialize(self, transport: "RESTServer", config: Dict[str, common.UrlMethod]) -> None:
        self._transport: "RESTServer" = transport
        self._config = config

    def _get_config(self, http_method: str) -> common.UrlMethod:
        if http_method.upper() not in self._config:
            allowed = ", ".join(self._config.keys())
            self.set_header("Allow", allowed)
            raise exceptions.BaseException(
                405, "%s is not supported for this url. Supported methods: %s" % (http_method, allowed)
            )

        return self._config[http_method]

    def get_auth_token(self, headers: Dict[str, str]) -> Optional[Dict[str, str]]:
        """
            Get the auth token provided by the caller. The token is provided as a bearer token.
        """
        if "Authorization" not in headers:
            return None

        parts = headers["Authorization"].split(" ")
        if len(parts) == 0 or parts[0].lower() != "bearer" or len(parts) > 2 or len(parts) == 1:
            LOGGER.warning(
                "Invalid authentication header, Inmanta expects a bearer token. (%s was provided)", headers["Authorization"]
            )
            return None

        return common.decode_token(parts[1])

    def respond(self, body: Optional[Dict[str, Any]], headers: Dict[str, str], status: int) -> None:
        if CONTENT_TYPE not in headers:
            headers[CONTENT_TYPE] = JSON_CONTENT
        elif headers[CONTENT_TYPE] != JSON_CONTENT:
            raise Exception("Invalid content type header provided. Only %s is supported" % JSON_CONTENT)

        if body is not None:
            self.write(common.json_encode(body))

        for header, value in headers.items():
            self.set_header(header, value)

        self.set_status(status)

    @gen.coroutine
    def _call(self, kwargs: Dict[str, str], http_method: str, call_config: common.UrlMethod) -> NoneGen:
        """
            An rpc like call
        """
        if call_config is None:
            raise exceptions.NotFound("This method does not exist")

        self.set_header("Access-Control-Allow-Origin", "*")
        try:
            message = self._transport._decode(self.request.body)
            if message is None:
                message = {}

            for key, value in self.request.query_arguments.items():
                if len(value) == 1:
                    message[key] = value[0].decode("latin-1")
                else:
                    message[key] = [v.decode("latin-1") for v in value]

            request_headers = self.request.headers
            auth_token = self.get_auth_token(request_headers)

            auth_enabled: bool = inmanta_config.Config.get("server", "auth", False)
            if not auth_enabled or auth_token is not None:
                result = yield self._transport._execute_call(
                    kwargs, http_method, call_config, message, request_headers, auth_token
                )
                self.respond(result.body, result.headers, result.status_code)
            else:
                raise exceptions.UnauthorizedException("Access to this resource is unauthorized.")

        except ValueError:
            LOGGER.exception("An exception occured")
            self.respond({"message": "Unable to decode request body"}, {}, 400)

        except exceptions.BaseException as e:
            self.respond(e.to_body(), {}, e.to_status())

        finally:
            self.finish()

    @gen.coroutine
    def head(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="HEAD", call_config=self._get_config("HEAD"), kwargs=kwargs)

    @gen.coroutine
    def get(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="GET", call_config=self._get_config("GET"), kwargs=kwargs)

    @gen.coroutine
    def post(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="POST", call_config=self._get_config("POST"), kwargs=kwargs)

    @gen.coroutine
    def delete(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="DELETE", call_config=self._get_config("DELETE"), kwargs=kwargs)

    @gen.coroutine
    def patch(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="PATCH", call_config=self._get_config("PATCH"), kwargs=kwargs)

    @gen.coroutine
    def put(self, *args: str, **kwargs: str) -> NoneGen:
        if args:
            raise Exception("Only named groups are support in url patterns")
        yield self._call(http_method="PUT", call_config=self._get_config("PUT"), kwargs=kwargs)

    def options(self, *args: str, **kwargs: str) -> None:
        if args:
            raise Exception("Only named groups are support in url patterns")

        allow_headers = "Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token, %s" % const.INMANTA_MT_HEADER
        if len(self._transport.headers):
            allow_headers += ", " + ", ".join(self._transport.headers)

        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "HEAD, GET, POST, PUT, OPTIONS, DELETE, PATCH")
        self.set_header("Access-Control-Allow-Headers", allow_headers)

        self.set_status(200)


class StaticContentHandler(tornado.web.RequestHandler):
    def initialize(self, transport: "RESTServer", content: str, content_type: str) -> None:
        self._transport = transport
        self._content = content
        self._content_type = content_type

    def get(self, *args: List[str], **kwargs: Dict[str, str]) -> None:
        self.set_header("Content-Type", self._content_type)
        self.write(self._content)
        self.set_status(200)


class RESTServer(RESTBase):
    """
        A tornado based rest server
    """

    _http_server: httpserver.HTTPServer

    def __init__(self, session_manager: common.SessionManagerInterface, id: str) -> None:
        super().__init__()

        self._id = id
        self.headers: Dict[str, str] = {}
        self.session_manager = session_manager

    def validate_sid(self, sid: uuid.UUID) -> bool:
        return self.session_manager.validate_sid(sid)

    @gen.coroutine
    def start(self, targets: List[inmanta.protocol.endpoints.CallTarget], additional_rules: List[routing.Rule] = []) -> NoneGen:
        """
            Start the server on the current ioloop
        """
        rules: List[routing.Rule] = []
        rules.extend(additional_rules)

        for slice in targets:
            url_map = slice.get_op_mapping()

            for url, configs in url_map.items():
                handler_config = {}
                for op, cfg in configs.items():
                    handler_config[op] = cfg

                rules.append(routing.Rule(routing.PathMatches(url), RESTHandler, {"transport": self, "config": handler_config}))
                LOGGER.debug("Registering handler(s) for url %s and methods %s" % (url, ", ".join(handler_config.keys())))

        port = 8888
        if self.id in inmanta_config.Config.get() and "port" in inmanta_config.Config.get()[self.id]:
            port = inmanta_config.Config.get()[self.id]["port"]

        application = web.Application(rules, compress_response=True)

        crt = inmanta_config.Config.get("server", "ssl_cert_file", None)
        key = inmanta_config.Config.get("server", "ssl_key_file", None)

        if crt is not None and key is not None:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(crt, key)

            self._http_server = httpserver.HTTPServer(application, decompress_request=True, ssl_options=ssl_ctx)
            LOGGER.debug("Created REST transport with SSL")
        else:
            self._http_server = httpserver.HTTPServer(application, decompress_request=True)
        self._http_server.listen(port)

        LOGGER.debug("Start REST transport")

    @gen.coroutine
    def stop(self) -> NoneGen:
        """
            Stop the current server
        """
        LOGGER.debug("Stopping Server Rest Endpoint")
        if self._http_server is not None:
            self._http_server.stop()

        yield self._http_server.close_all_connections()



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from tornado import web
from typing import Optional, Dict, Any


class BaseException(web.HTTPError):
    """
        A base exception for errors in the server
    """

    def __init__(self, status_code: int = 500, message: Optional[str] = None) -> None:
        super().__init__(status_code, message)

    def to_body(self) -> Dict[str, Any]:
        """
            Return a response body
        """
        return {"message": self.log_message}

    def to_status(self) -> int:
        """
            Return the status code
        """
        return self.status_code


class AccessDeniedException(BaseException):
    """
        An exception raised when access is denied (403)
    """

    def __init__(self, message: Optional[str] = None) -> None:
        msg = "Access denied"
        if message is not None:
            msg += ": " + message

        super().__init__(403, msg)


class UnauthorizedException(BaseException):
    """
        An exception raised when access to this resource is unauthorized
    """

    def __init__(self, message: Optional[str] = None) -> None:
        msg = "Access to this resource is unauthorized"
        if message is not None:
            msg += ": " + message

        super().__init__(401, msg)


class BadRequest(BaseException):
    """
        This exception is raised for a mailformed request
    """

    def __init__(self, message: Optional[str] = None) -> None:
        msg = "Invalid request"
        if message is not None:
            msg += ": " + message

        super().__init__(400, msg)


class NotFound(BaseException):
    """
        This exception is used to indicate that a request or reference resource was not found.
    """

    def __init__(self, message: Optional[str] = None) -> None:
        msg = "Request or referenced resource does not exist"
        if message is not None:
            msg += ": " + message

        super().__init__(404, msg)


class ServerError(BaseException):
    """
        An unexpected error occurred in the server
    """

    def __init__(self, message: Optional[str] = None) -> None:
        msg = "An unexpected error occurred in the server while processing the request"
        if message is not None:
            msg += ": " + message

        super().__init__(500, msg)



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import pytest

from inmanta.ast import RuntimeException
import inmanta.compiler as compiler


def test_option_values(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test1:

end

entity Test2:
    bool flag=false
end

implement Test2 using std::none

Test1 test1 [1] -- [0:1] Test2 other

implementation tt for Test1:

end

implement Test1 using tt when self.other.flag == false

Test1()
"""
    )
    with pytest.raises(RuntimeException):
        compiler.do_compile()


def test_isset(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
entity Test1:

end

entity Test2:
    bool flag=false
end

implement Test2 using std::none

Test1 test1 [1] -- [0:1] Test2 other

implementation tt for Test1:

end

implement Test1 using tt when self.other is defined and self.other.flag == false

Test1(other=Test2())
"""
    )
    compiler.do_compile()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import threading
import logging
import time

from execnet import multi, gateway_bootstrap
from . import local
from inmanta import resources


LOGGER = logging.getLogger()


class CannotLoginException(Exception):
    pass


class RemoteException(Exception):
    def __init__(self, exception_type, msg, traceback=None):
        super().__init__(exception_type, msg, traceback)


class SshIO(local.IOBase):
    """
        This class provides handler IO methods. This io method is used when the ssh scheme is provided in the agent uri.

        The uri supports setting the hostname, user and port. In the query string the following config can be provided:
         * python: The python interpreter to use. The default value is python
         * retries: The number of retries before giving up. The default number of retries 10
         * retry_wait: The time to wait between retries for the remote target to become available. The default wait is 30s
    """
    def is_remote(self):
        return True

    def __init__(self, uri, config):
        super(SshIO, self).__init__(uri, config)
        self._host = config["host"]
        if "port" in config and config["port"] is not None:
            self._port = int(config["port"])
        else:
            self._port = 22

        if "user" in config and config["user"] is not None:
            self._user = config["user"]
        else:
            self._user = "root"

        if "retries" in config and config["retries"] is not None:
            self._retries = int(config["retries"])
        else:
            self._retries = 10

        if "retry_wait" in config and config["retry_wait"] is not None:
            self._retry_wait = int(config["retry_wait"])
        else:
            self._retry_wait = 30

        self._lock = threading.Lock()
        self._group = multi.Group()
        self._gw = None
        connect = self._build_connect_string()
        LOGGER.info("Starting execnet connection group %s", id(self._group))
        try:
            attempts = self._retries + 1
            while attempts > 0:
                try:
                    self._gw = self._group.makegateway(self._build_connect_string())
                    attempts = 0
                except gateway_bootstrap.HostNotFound as e:
                    attempts -= 1
                    if attempts == 0:
                        self._group.terminate(0.1)
                        raise resources.HostNotFoundException(hostname=self._host, user=self._user, error=e)

                    LOGGER.info("Failed to login to %s, waiting %d seconds and %d attempts left.",
                                self.uri, self._retry_wait, attempts)
                    time.sleep(self._retry_wait)

        except BrokenPipeError as e:
            LOGGER.info("Terminating execnet connection group due to exception %s", id(self._group), e)
            self._group.terminate(0.1)
            raise resources.HostNotFoundException(hostname=self._host, user=self._user, error=e)
        except AssertionError as e:
            LOGGER.info("Terminating execnet connection group due to exception %s", id(self._group), e)
            self._group.terminate(0.1)
            raise CannotLoginException()

        assert self._gw is not None
        LOGGER.info("Connected with %s", connect)

    def _build_connect_string(self):
        """
            Build the connection string for execent based on the hostname
        """
        opts = "-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -o PasswordAuthentication=no"
        opts += " -p %d" % self._port

        if "python" in self.config:
            python = self.config["python"]
        else:
            python = "python"

        return "ssh=%s %s@%s//python=%s" % (opts, self._user, self._host, python)

    def _execute(self, function_name, *args, **kwargs):
        with self._lock:
            ch = self._gw.remote_exec(local)
            ch.send((function_name, args, kwargs))
            result = ch.receive()
            ch.close()

        # check if we got an exception
        if isinstance(result, dict) and "__type__" in result and result["__type__"] == "RemoteException":
            raise RemoteException(exception_type=result["exception_type"], msg=result["exception_string"],
                                  traceback=result["traceback"])

        return result

    def read_binary(self, path):
        # remoting can turn this into a string
        result = self._execute("read_binary", path)
        if isinstance(result, str):
            return result.encode()
        return result

    def __getattr__(self, name):
        """
            Proxy a function call to the local version on the other side of the channel.
        """
        def call(*args, **kwargs):
            result = self._execute(name, *args, **kwargs)
            return result

        return call

    def close(self):
        LOGGER.info("Terminating execnet connection group %s", id(self._group))
        if self._group is not None:
            self._group.terminate(0.1)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from copy import copy
from collections import Mapping

from inmanta.execute.util import Unknown, NoneValue
from inmanta.ast import RuntimeException


class UnsetException(RuntimeException):
    """
        This exception is thrown when an attribute is accessed that was not yet
        available (i.e. it has not been frozen yet).
    """

    def __init__(self, msg, instance=None, attribute=None):
        RuntimeException.__init__(self, None, msg)
        self.instance = instance
        self.attribute = attribute
        self.msg = msg

    def get_result_variable(self):
        return self.instance


class UnknownException(Exception):
    """
        This exception is thrown when code tries to access a value that is
        unknown and cannot be determined during this evaluation. The code
        receiving this exception is responsible for invalidating any results
        depending on this value by return an instance of Unknown as well.
    """

    def __init__(self, unknown):
        self.unknown = unknown


class DynamicProxy(object):
    """
        This class wraps an object and makes sure that a model is never modified
        by native code.
    """

    def __init__(self, instance):
        object.__setattr__(self, "__instance", instance)

    def _get_instance(self):
        return object.__getattribute__(self, "__instance")

    @classmethod
    def unwrap(cls, item):
        if isinstance(item, DynamicProxy):
            return item._get_instance()

        if isinstance(item, list):
            return [cls.unwrap(x) for x in item]

        return item

    @classmethod
    def return_value(cls, value):
        if value is None:
            return None

        if isinstance(value, NoneValue):
            return None

        if isinstance(value, Unknown):
            raise UnknownException(value)

        if isinstance(value, (str, tuple, int, float, bool)):
            return copy(value)

        if isinstance(value, DynamicProxy):
            return value

        if isinstance(value, dict):
            return DictProxy(value)

        if hasattr(value, "__len__"):
            return SequenceProxy(value)

        if hasattr(value, "__call__"):
            return CallProxy(value)

        return DynamicProxy(value)

    def __getattr__(self, attribute):
        instance = self._get_instance()
        value = instance.get_attribute(attribute).get_value()

        return DynamicProxy.return_value(value)

    def __setattr__(self, attribute, value):
        raise Exception("Readonly object")

    def _type(self):
        """
            Return the type of the proxied instance
        """
        return self._get_instance().type

    def is_unknown(self):
        """
            Return true if this value is unknown and cannot be determined
            during this compilation run
        """
        if isinstance(self._get_instance(), Unknown):
            return True
        return False

    def __hash__(self):
        return hash(self._get_instance())

    def __eq__(self, other):
        if hasattr(other, "_get_instance"):
            other = other._get_instance()

        return self._get_instance() == other

    def __lt__(self, other):
        if hasattr(other, "_get_instance"):
            other = other._get_instance()

        return self._get_instance() < other

    def __repr__(self):
        return "@%s" % repr(self._get_instance())


class SequenceProxy(DynamicProxy):

    def __init__(self, iterator):
        DynamicProxy.__init__(self, iterator)

    def __getitem__(self, key):
        instance = self._get_instance()
        if isinstance(key, str):
            raise RuntimeException(self, "can not get a attribute %s, %s is a list" % (key, self._get_instance()))

        return DynamicProxy.return_value(instance[key])

    def __len__(self):
        return len(self._get_instance())

    def __iter__(self):
        instance = self._get_instance()

        return IteratorProxy(instance.__iter__())


class DictProxy(DynamicProxy, Mapping):

    def __init__(self, mydict):
        DynamicProxy.__init__(self, mydict)

    def __getitem__(self, key):
        instance = self._get_instance()
        if not isinstance(key, str):
            raise RuntimeException(self, "Expected string key, but got %s, %s is a dict" % (key, self._get_instance()))

        return DynamicProxy.return_value(instance[key])

    def __len__(self):
        return len(self._get_instance())

    def __iter__(self):
        instance = self._get_instance()

        return IteratorProxy(instance.__iter__())


class CallProxy(DynamicProxy):
    """
        Proxy a value that implements a __call__ function
    """

    def __init__(self, instance):
        DynamicProxy.__init__(self, instance)

    def __call__(self, *args, **kwargs):
        instance = self._get_instance()

        return instance(*args, **kwargs)


class IteratorProxy(DynamicProxy):
    """
        Proxy an iterator call
    """

    def __init__(self, iterator):
        DynamicProxy.__init__(self, iterator)

    def __iter__(self):
        return self

    def __next__(self):
        i = self._get_instance()
        return DynamicProxy.return_value(next(i))



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import time

from inmanta.ast.statements import DefinitionStatement, TypeDefinitionStatement
from inmanta.const import LOG_LEVEL_TRACE
from inmanta.execute.proxy import UnsetException
from inmanta import plugins
from inmanta.ast.type import TYPES, Type

from inmanta.ast.statements.define import DefineEntity, DefineImplement, DefineTypeDefault, DefineIndex, DefineRelation
from inmanta.execute.runtime import Resolver, ExecutionContext, QueueScheduler, ExecutionUnit
from inmanta.ast.entity import Entity
from inmanta.ast import RuntimeException, MultiException, CycleExcpetion
from inmanta.execute.tracking import ModuleTracker
import itertools
from typing import Dict, List, Set

DEBUG = True
LOGGER = logging.getLogger(__name__)

MAX_ITERATIONS = 2000


class Scheduler(object):
    """
        This class schedules statements for execution
    """

    def __init__(self):
        pass

    def freeze_all(self, exns):
        for t in [t for t in self.types.values() if isinstance(t, Entity)]:
            t.final(exns)

        instances = self.types["std::Entity"].get_all_instances()

        for i in instances:
            i.final(exns)

    def dump(self, type="std::Entity"):
        instances = self.types[type].get_all_instances()

        for i in instances:
            i.dump()

    def verify_done(self):
        instances = self.types["std::Entity"].get_all_instances()
        notdone = []
        for i in instances:
            if not i.verify_done():
                notdone.append(i)

        return notdone

    def get_types(self):
        return self.types

    def get_scopes(self):
        return self.scopes

    def dump_not_done(self):

        for i in self.verify_done():
            i.dump()

    def sort_entities(self, entity_map: Dict[str, DefineEntity]) -> List[DefineEntity]:
        out: List[DefineEntity] = []
        loopstack: Set[str] = set()
        while len(entity_map) > 0:
            workon = next(iter(entity_map.keys()))
            self.do_sort_entities(entity_map, workon, out, loopstack)
        return out

    def do_sort_entities(self,
                         entity_map: Dict[str, DefineEntity],
                         name: str,
                         acc: List[DefineEntity],
                         loopstack: Set[str]) -> None:
        nexte = entity_map[name]
        try:
            del entity_map[name]
            loopstack.add(name)
            for p in nexte.get_full_parent_names():
                if p in loopstack:
                    raise CycleExcpetion(nexte, p)
                if p in entity_map:
                    self.do_sort_entities(entity_map, p, acc, loopstack)
            loopstack.remove(name)
            acc.append(nexte)
        except CycleExcpetion as ce:
            ce.add(nexte)
            raise

    def define_types(self, compiler, statements, blocks):
        """
            This is the first compiler stage that defines all types_and_impl
        """
        # get all relevant stmts
        definitions = [d for d in statements if isinstance(d, DefinitionStatement)]
        others = [d for d in statements if not isinstance(d, DefinitionStatement)]

        if not len(others) == 0:
            raise Exception("others not empty %s" % repr(others))

        # collect all  types and impls
        types_and_impl = {}

        # set primitive types
        compiler.get_ns().set_primitives(TYPES)

        # all stmts contributing types and impls
        newtypes = [k for k in [t.register_types()
                                for t in definitions if isinstance(t, TypeDefinitionStatement)] if k is not None]

        for (name, type_symbol) in newtypes:
            types_and_impl[name] = type_symbol

        # now that we have objects for all types, populate them
        implements = [t for t in definitions if isinstance(t, DefineImplement)]
        others = [t for t in definitions if not isinstance(t, DefineImplement)]
        entities = {t.fullName: t for t in others if isinstance(t, DefineEntity)}
        typedefaults = [t for t in others if isinstance(t, DefineTypeDefault)]
        others = [t for t in others if not (isinstance(t, DefineEntity) or isinstance(t, DefineTypeDefault))]
        indices = [t for t in others if isinstance(t, DefineIndex)]
        others = [t for t in others if not isinstance(t, DefineIndex)]

        # first entities, so we have inheritance
        # parents first
        for d in self.sort_entities(entities):
            d.evaluate()

        for d in typedefaults:
            d.evaluate()

        for d in others:
            d.evaluate()

        # indices late, as they require all attributes
        for d in indices:
            d.evaluate()

        # lastly the implements, as they require implementations
        for d in implements:
            d.evaluate()

        types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type) or isinstance(v, plugins.Plugin)}
        compiler.plugins = {k: v for k, v in types_and_impl.items() if isinstance(v, plugins.Plugin)}

        # give type info to all types, to normalize blocks inside them
        for t in types.values():
            t.normalize()

        # normalize root blocks
        for block in blocks:
            block.normalize()

        self.types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type)}

    def anchormap(self, compiler, statements, blocks):
        prev = time.time()

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)

        # relations are also in blocks
        statements = (s for s in statements if not isinstance(s, DefineRelation))
        anchors = (anchor for container in itertools.chain(statements, blocks)
                   for anchor in container.get_anchors() if anchor is not None)

        rangetorange = [(anchor.get_location(), anchor.resolve()) for anchor in anchors]
        rangetorange = [(f, t) for f, t in rangetorange if t is not None]

        now = time.time()
        LOGGER.debug("Anchormap took %f seconds", now - prev)

        return rangetorange

    def run(self, compiler, statements, blocks):
        """
            Evaluate the current graph
        """
        prev = time.time()

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)

        # give all loose blocks an empty XC
        # register the XC's as scopes
        # All named scopes are now present
        for block in blocks:
            res = Resolver(block.namespace)
            xc = ExecutionContext(block, res)
            block.context = xc
            block.namespace.scope = xc

        # setup queues
        # queue for runnable items
        basequeue = []
        # queue for RV's that are delayed
        waitqueue = []
        # queue for RV's that are delayed and had no effective waiters when they were first in the waitqueue
        zerowaiters = []
        # queue containing everything, to find hanging statements
        all_statements = []

        # Wrap in object to pass around
        queue = QueueScheduler(compiler, basequeue, waitqueue, self.types, all_statements)

        # emit all top level statements
        for block in blocks:
            block.context.emit(queue.for_tracker(ModuleTracker(block)))

        # start an evaluation loop
        i = 0
        count = 0
        while i < MAX_ITERATIONS:
            now = time.time()

            # check if we can stop the execution
            if len(basequeue) == 0 and len(waitqueue) == 0 and len(zerowaiters) == 0:
                break
            else:
                i += 1

            LOGGER.debug("Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)", i,
                         len(basequeue), len(waitqueue), len(zerowaiters), count, now - prev)
            prev = now

            # evaluate all that is ready
            while len(basequeue) > 0:
                next = basequeue.pop(0)
                try:
                    next.execute()
                    count = count + 1
                except UnsetException as e:
                    # some statements don't know all their dependencies up front,...
                    next.waitfor(e.get_result_variable())

            # all safe stmts are done
            progress = False

            # find a RV that has waiters, so freezing creates progress
            while len(waitqueue) > 0 and not progress:
                next = waitqueue.pop(0)
                if next.get_progress_potential() == 0:
                    zerowaiters.append(next)
                elif next.get_waiting_providers() > 0:
                    # definitely not done
                    # drop from queue
                    # will requeue when value is added
                    next.unqueue()
                else:
                    # freeze it and go to next iteration, new statements will be on the basequeue
                    LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next)
                    next.freeze()
                    progress = True

            # no waiters in waitqueue,...
            # see if any zerowaiters have become gotten waiters
            if not progress:
                waitqueue = [w for w in zerowaiters if w.get_progress_potential() != 0]
                queue.waitqueue = waitqueue
                zerowaiters = [w for w in zerowaiters if w.get_progress_potential() == 0]
                while len(waitqueue) > 0 and not progress:
                    LOGGER.debug("Moved zerowaiters to waiters")
                    next = waitqueue.pop(0)
                    if next.get_waiting_providers() > 0:
                        next.unqueue()
                    else:
                        LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next)
                        next.freeze()
                        progress = True

            # no one waiting anymore, all done, freeze and finish
            if not progress:
                LOGGER.debug("Finishing statements with no waiters")
                while len(zerowaiters) > 0:
                    next = zerowaiters.pop()
                    next.freeze()

        now = time.time()
        LOGGER.debug("Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)", i,
                     len(basequeue), len(waitqueue), len(zerowaiters), count, now - prev)

        if i == MAX_ITERATIONS:
            print("could not complete model")
            return False
        # now = time.time()
        # print(now - prev)
        # end evaluation loop
        # self.dump_not_done()
        # print(basequeue, waitqueue)
        # dumpHangs()
        # self.dump()
        # rint(len(self.types["std::Entity"].get_all_instances()))

        excns = []
        self.freeze_all(excns)

        if len(excns) == 0:
            pass
        elif len(excns) == 1:
            raise excns[0]
        else:
            raise MultiException(excns)

        all_statements = [x for x in all_statements if not x.done]

        if all_statements:
            stmt = None
            for st in all_statements:
                if isinstance(st, ExecutionUnit):
                    stmt = st
                    break

            raise RuntimeException(stmt.expression, "not all statements executed %s" % all_statements)
        # self.dump("std::File")

        return True



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import sys
import os
import subprocess
import pytest
import re
import conftest
from threading import Timer
from inmanta import const
import signal
from subprocess import TimeoutExpired


def get_command(tmp_dir, stdout_log_level=None, log_file=None, log_level_log_file=None, timed=False):
    root_dir = tmp_dir.mkdir("root").strpath
    log_dir = os.path.join(root_dir, "log")
    state_dir = os.path.join(root_dir, "data")
    for directory in [log_dir, state_dir]:
        os.mkdir(directory)
    config_file = os.path.join(root_dir, "inmanta.cfg")

    port = conftest.get_free_tcp_port()

    with open(config_file, 'w+') as f:
        f.write("[config]\n")
        f.write("log-dir=" + log_dir + "\n")
        f.write("state-dir=" + state_dir + "\n")
        f.write("[database]\n")
        f.write("port=" + str(port) + "\n")

    args = [sys.executable, "-m", "inmanta.app"]
    if stdout_log_level:
        args.append("-" + "v" * stdout_log_level)
    if log_file:
        log_file = os.path.join(log_dir, log_file)
        args += ["--log-file", log_file]
    if log_file and log_level_log_file:
        args += ["--log-file-level", str(log_level_log_file)]
    if timed:
        args += ["--timed-logs"]
    args += ["-c", config_file, "server"]
    return (args, log_dir)


def do_run(args, env={}):
    baseenv = os.environ.copy()
    baseenv.update(env)
    process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=baseenv)
    return process


def convert_to_ascii(text):
    return [line for line in text.decode("ascii").split("\n") if line != ""]


def do_kill(process, killtime=3, termtime=2):
    t1 = Timer(killtime, process.kill)
    t2 = Timer(termtime, process.terminate)
    t1.start()
    t2.start()

    out, err = process.communicate()

    print(process.returncode)

    t1.cancel()
    t2.cancel()

    stdout = convert_to_ascii(out)
    stderr = convert_to_ascii(err)
    return (stdout, stderr)


def run_without_tty(args, env={}, killtime=3, termtime=2):
    process = do_run(args, env)
    return do_kill(process, killtime, termtime)


def run_with_tty(args):
    """Could not get code for actual tty to run stable in docker, so we are faking it """
    env = {const.ENVIRON_FORCE_TTY: "true"}
    return run_without_tty(args, env=env)


def get_timestamp_regex():
    return r'[\d]{4}\-[\d]{2}\-[\d]{2} [\d]{2}\:[\d]{2}\:[\d]{2}\,[\d]{3}'


def get_compiled_regexes(regexes, timed):
    result = []
    for regex in regexes:
        if timed:
            regex = get_timestamp_regex() + ' ' + regex
        compiled_regex = re.compile(regex)
        result.append(compiled_regex)
    return result


def is_colorama_package_available():
    try:
        import colorama  # noqa: F401
    except ModuleNotFoundError:
        return False
    return True


def test_verify_that_colorama_package_is_not_present():
    """
        The colorama package turns the colored characters in TTY-based terminal into uncolored characters.
        As such, this package should not be present.
    """
    assert not is_colorama_package_available()


@pytest.mark.parametrize("log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines", [
    (3, False, False, [r'[a-z.]*[ ]*INFO[\s]+Starting server endpoint',
                       r'[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint'], []),
    (2, False, False, [r'[a-z.]*[ ]*INFO[\s]+Starting server endpoint'],
                      [r'[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint']),
    (3, False, True, [r'\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m \x1b\[34mStarting server endpoint',
                      r'\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m \x1b\[34mStarting Server Rest Endpoint'], []),
    (2, False, True, [r'\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m \x1b\[34mStarting server endpoint'],
                     [r'\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m \x1b\[34mStarting Server Rest Endpoint']),
    (3, True, False, [r'[a-z.]*[ ]*INFO[\s]+Starting server endpoint',
                      r'[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint'], []),
    (2, True, False, [r'[a-z.]*[ ]*INFO[\s]+Starting server endpoint'],
                     [r'[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint']),
    (3, True, True, [r'\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m \x1b\[34mStarting server endpoint',
                     r'\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m \x1b\[34mStarting Server Rest Endpoint'], []),
    (2, True, True, [r'\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m \x1b\[34mStarting server endpoint'],
                    [r'\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m \x1b\[34mStarting Server Rest Endpoint'])
])
@pytest.mark.timeout(20)
def test_no_log_file_set(tmpdir, log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines):
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    (args, log_dir) = get_command(tmpdir, stdout_log_level=log_level, timed=timed)
    if with_tty:
        (stdout, _) = run_with_tty(args)
    else:
        (stdout, _) = run_without_tty(args)
    log_file = "server.log"
    assert log_file not in os.listdir(log_dir)
    assert len(stdout) != 0
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, timed)


@pytest.mark.parametrize("log_level, with_tty, regexes_required_lines, regexes_forbidden_lines", [
    (3, False, [r'[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint',
                r'[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint'], []),
    (2, False, [r'[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint'],
     [r'[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint']),
    (3, True, [r'[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint',
               r'[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint'], []),
    (2, True, [r'[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint'],
     [r'[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint'])
])
@pytest.mark.timeout(60)
def test_log_file_set(tmpdir, log_level, with_tty, regexes_required_lines, regexes_forbidden_lines):
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    log_file = "server.log"
    (args, log_dir) = get_command(tmpdir, stdout_log_level=log_level, log_file=log_file, log_level_log_file=log_level)
    if with_tty:
        (stdout, _) = run_with_tty(args)
    else:
        (stdout, _) = run_without_tty(args)
    assert log_file in os.listdir(log_dir)
    log_file = os.path.join(log_dir, log_file)
    with open(log_file, 'r') as f:
        log_lines = f.readlines()
    check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed=True)
    check_logs(stdout, [], regexes_required_lines, timed=True)
    check_logs(stdout, [], regexes_required_lines, timed=False)


def check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed):
    compiled_regexes_requires_lines = get_compiled_regexes(regexes_required_lines, timed)
    compiled_regexes_forbidden_lines = get_compiled_regexes(regexes_forbidden_lines, timed)
    for line in log_lines:
        print(line)
    for regex in compiled_regexes_requires_lines:
        if not any(regex.match(line) for line in log_lines):
            pytest.fail("Required pattern was not found in log lines: %s" % (regex.pattern,))
    for regex in compiled_regexes_forbidden_lines:
        if any(regex.match(line) for line in log_lines):
            pytest.fail("Forbidden pattern found in log lines: %s" % (regex.pattern,))


def test_check_shutdown():
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py")])
    # wait for handler to be in place
    try:
        process.communicate(timeout=1)
    except TimeoutExpired:
        pass
    process.send_signal(signal.SIGUSR1)
    out, err = do_kill(process, killtime=5, termtime=2)
    print(out, err)
    assert "----- Thread Dump ----" in out
    assert "STOP" in out
    assert "SHUTDOWN COMPLETE" in out


def test_check_bad_shutdown():
    print([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    out, err = do_kill(process, killtime=5, termtime=2)
    print(out, err)
    assert "----- Thread Dump ----" in out
    assert "STOP" not in out
    assert "SHUTDOWN COMPLETE" not in out



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging

from inmanta.execute.runtime import ResultVariable, ExecutionUnit, RawUnit, HangUnit, Instance, Resolver, QueueScheduler,\
    ResultCollector
from inmanta.ast.statements.assign import Assign, SetAttribute
from inmanta.ast.statements import ExpressionStatement, AssignStatement
from inmanta.ast import RuntimeException, Locatable, Location, LocatableString
from typing import List, Dict
from inmanta.parser import ParserException

LOGGER = logging.getLogger(__name__)


class Reference(ExpressionStatement):
    """
        This class represents a reference to a value
    """

    def __init__(self, name: LocatableString) -> None:
        ExpressionStatement.__init__(self)
        self.name = str(name)
        self.full_name = str(name)

    def normalize(self) -> None:
        pass

    def requires(self) -> List[str]:
        return [self.full_name]

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, ResultVariable]:
        # FIXME: may be done more efficient?
        out = {self.name: resolver.lookup(self.full_name)}  # type : Dict[object, ResultVariable]
        return out

    def requires_emit_gradual(self, resolver: Resolver, queue: QueueScheduler, resultcollector) -> Dict[object, ResultVariable]:
        var = resolver.lookup(self.full_name)
        var.listener(resultcollector, self.location)
        out = {self.name: var}  # type : Dict[object, ResultVariable]
        return out

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        return requires[self.name]

    def execute_direct(self, requires: Dict[object, object]) -> object:
        if self.name not in requires:
            raise RuntimeException(self, "Could not resolve the value %s in this static context" % self.name)
        return requires[self.name]

    def as_assign(self, value: ExpressionStatement, list_only: bool=False) -> AssignStatement:
        if list_only:
            raise ParserException(self.location, "+=", "Can not perform += on variable %s" % self.name)
        return Assign(self.name, value)

    def root_in_self(self) -> "Reference":
        if self.name == 'self':
            return self
        else:
            ref = Reference('self')
            self.copy_location(ref)
            attr_ref = AttributeReference(ref, self.name)
            self.copy_location(attr_ref)
            return attr_ref

    def __str__(self) -> str:
        return self.name

    def __repr__(self, *args, **kwargs):
        return self.name


class AttributeReferenceHelper(Locatable):
    """
        Helper class for AttributeReference, reschedules itself
    """

    def __init__(self, target: ResultVariable, instance: Reference, attribute: str, resultcollector: ResultCollector) -> None:
        Locatable.__init__(self)
        self.attribute = attribute
        self.target = target
        self.instance = instance
        self.resultcollector = resultcollector

    def resume(self,
               requires: Dict[object, object],
               resolver: Resolver,
               queue_scheduler: QueueScheduler,
               target: ResultVariable) -> None:
        """
            Instance is ready to execute, do it and see if the attribute is already present
        """
        # get the Instance
        obj = self.instance.execute(requires, resolver, queue_scheduler)

        if isinstance(obj, list):
            raise RuntimeException(self, "can not get a attribute %s, %s is a list" % (self.attribute, obj))

        if not isinstance(obj, Instance):
            raise RuntimeException(self, "can not get a attribute %s, %s not an entity" % (self.attribute, obj))

        # get the attribute result variable
        attr = obj.get_attribute(self.attribute)
        # Cache it
        self.attr = attr

        if attr.is_ready():
            # go ahead
            # i.e. back to the AttributeReference itself
            self.target.set_value(attr.get_value(), self.location)
        else:
            # reschedule on the attribute, XU will assign it to the target variable
            if self.resultcollector is not None:
                attr.listener(self.resultcollector, self.location)
            ExecutionUnit(queue_scheduler, resolver, self.target, {"x": attr}, self)

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        # Attribute is ready, return it,
        return self.attr.get_value()

    def __str__(self) -> str:
        return "%s.%s" % (self.instance, self.attribute)

    def get_location(self) -> Location:
        return self.location


class IsDefinedReferenceHelper(Locatable):
    """
        Helper class for AttributeReference, reschedules itself
    """

    def __init__(self, target: ResultVariable, instance: Reference, attribute: str) -> None:
        Locatable.__init__(self)
        self.attribute = attribute
        self.target = target
        self.instance = instance

    def resume(self, requires: Dict[object, ResultVariable], resolver: Resolver, queue_scheduler: QueueScheduler) -> None:
        """
            Instance is ready to execute, do it and see if the attribute is already present
        """
        try:
            # get the Instance
            obj = self.instance.execute({k: v.get_value() for k, v in requires.items()}, resolver, queue_scheduler)

            if isinstance(obj, list):
                raise RuntimeException(self, "can not get a attribute %s, %s is a list" % (self.attribute, obj))

            # get the attribute result variable
            attr = obj.get_attribute(self.attribute)
            # Cache it
            self.attr = attr

            if attr.is_ready():
                # go ahead
                # i.e. back to the AttributeReference itself
                attr.get_value()
                self.target.set_value(True, self.location)
            else:
                requires["x"] = attr
                # reschedule on the attribute, XU will assign it to the target variable
                RawUnit(queue_scheduler, resolver, requires, self)

        except RuntimeException:
            self.target.set_value(False, self.location)

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue_scheduler: QueueScheduler) -> object:
        # Attribute is ready, return it,
        return self.attr.get_value()


class AttributeReference(Reference):
    """
        This variable refers to an attribute. This is mostly used to refer to
        attributes of a class or class instance.
    """

    def __init__(self, instance: Reference, attribute: LocatableString) -> None:
        Reference.__init__(self, "%s.%s" % (instance.full_name, attribute))
        self.attribute = str(attribute)

        # a reference to the instance
        self.instance = instance

    def requires(self) -> List[str]:
        return self.instance.requires()

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, ResultVariable]:
        return self.requires_emit_gradual(resolver, queue, None)

    def requires_emit_gradual(self, resolver: Resolver, queue: QueueScheduler, resultcollector) -> Dict[object, ResultVariable]:
        # The tricky one!

        # introduce temp variable to contain the eventual result of this stmt
        temp = ResultVariable()
        temp.set_provider(self)

        # construct waiter
        resumer = AttributeReferenceHelper(temp, self.instance, self.attribute, resultcollector)
        self.copy_location(resumer)

        # wait for the instance
        HangUnit(queue, resolver, self.instance.requires_emit(resolver, queue), None, resumer)
        return {self: temp}

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        # helper returned: return result
        return requires[self]

    def as_assign(self, value: ExpressionStatement, list_only: bool=False) -> AssignStatement:
        return SetAttribute(self.instance, self.attribute, value, list_only)

    def root_in_self(self) -> Reference:
        out = AttributeReference(self.instance.root_in_self(), self.attribute)
        self.copy_location(out)
        return out

    def __repr__(self, *args, **kwargs):
        return "%s.%s" % (repr(self.instance), self.attribute)
